from typing import List, Dict

from rb.core.block import Block
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.community import Community
from rb.cna.cna_graph import CnaGraph
from rb.core.cscl.regularity_indices import RegularityIndices

from rb.utils.rblogger import Logger
from bisect import bisect_left
from copy import deepcopy
import math
from scipy.special import entr
from numpy.linalg import norm
import numpy as np

L_d = None
L_m = None
L_w = None
user_list_map = None


def get_all_timestamps_for_user(given_user, community: Community):
    timestamps = []
    user_contribution = community.get_participant_contributions(given_user)
    for contribution in user_contribution:
        timestamp = contribution.get_timestamp()
        if timestamp not in timestamps:
            timestamps.append(timestamp)

    timestamps.sort()
    return timestamps


def get_user_list_map(community: Community):
    users = community.get_participants()
    for user in users:
        timestamps = get_all_timestamps_for_user(user, community)
        user_list_map[user] = timestamps


def F(user, W, x):
    user_list = user_list_map[user]
    to_search = W * x
    index = bisect_left(user_list, to_search)

    if index == len(user_list):
        return 0
    elif user_list[index] <= (W * x + W):
        return 1

    return 0


def D(user, h):
    total = 0

    for i in range(L_d):
        total += F(user, 60, 24 * i + h)

    return total


def W(user, d):
    total = 0

    for i in range(L_w):
        total += F(user, 60 * 24, 7 * i + d)

    return total


def compute_time_measures(user):
    user_d = []
    user_w = []

    for h in range(24):
        user_d.append(D(user, h))
    for d in range(7):
        user_w.append(W(user, d))

    max_d = max(user_d)
    max_w = max(user_w)
    total_d = sum(user_d)
    total_w = sum(user_w)
    normalised_d = deepcopy(user_d)
    normalised_w = deepcopy(user_w)

    if total_d != 0:
        normalised_d = list(map(lambda x: float(x) / total_d, user_d))
    if total_w != 0:
        normalised_w = list(map(lambda x: float(x) / total_w, user_w))

    e_d = 0
    for h in range(24):
        if normalised_d[h] != 0:
            e_d += (normalised_d[h] * math.log(normalised_d[h]))

    e_w = 0
    for d in range(7):
        if normalised_w[d] != 0:
            e_w += (normalised_w[d] * math.log(normalised_w[d]))

    PDH = (math.log(24) - e_d) * max_d
    PWD = (math.log(7) - e_w) * max_w

    return PDH, PWD


def P(user, d, k):
    total = 0

    for i in range(24):
        total += F(user, 60, 24 * (d + 7 * k) + i)

    return total


def get_profile(user, k):
    return [P(user, d, k) for d in range(7)]


def active(user, k):
    profile = get_profile(user, k)
    active_days = []

    for d in range(7):
        if profile[d] != 0:
            active_days.append(d)

    return active_days


def JSD(P, Q):
    _P, _Q = np.array(P), np.array(Q)
    sum_p = sum(P)
    sum_q = sum(Q)

    if sum_p != 0:
        _P = _P / sum_p
    if sum_q != 0:
        _Q = _Q / sum_q

    return entr(0.5 * (_P + _Q)).sum() - 0.5 * (entr(_P).sum() + entr(_Q).sum())


def similarity_1(user, i, j):
    active_i = active(user, i)
    active_j = active(user, j)
    active_common = list(set(active_i) & set(active_j))

    max_len = max(len(active_i), len(active_j))
    if max_len == 0:
        return 0

    return float(len(active_common)) / max_len


def similarity_2(user, i, j):
    profile_i = get_profile(user, i)
    profile_j = get_profile(user, j)

    return 1 - (JSD(profile_i, profile_j) / math.log(2))


def similarity_3(user, i, j):
    active_i = active(user, i)
    active_j = active(user, j)
    active_all = list(set(active_i) | set(active_j))

    if len(active_all) == 0:
        return 0

    total = 0
    for d in range(7):
        p_i = P(user, d, i)
        p_j = P(user, d, j)

        if (p_i + p_j) != 0:
            total += math.pow(float(p_i - p_j) / (p_i + p_j), 2)

    return 1 - float(total) / len(active_all)


def compute_similarities(user):
    ws1_total = 0
    ws2_total = 0
    ws3_total = 0
    count = 0

    for i in range(L_w):
        for j in range(i + 1, L_w):
            count += 1
            ws1_total += similarity_1(user, i, j)
            ws2_total += similarity_2(user, i, j)
            ws3_total += similarity_3(user, i, j)

    if count == 0:
        count = 1

    return float(ws1_total) / count, float(ws2_total) / count, float(ws3_total) / count


def determine_regularity(community: Community):
    global L_d
    global L_m
    global L_w
    users = community.get_participants()
    get_user_list_map(community)
    for i in range(0, len(users)):
        user_timestamps = user_list_map[users[i]]
        last_comment_timestamp = user_timestamps[len(user_timestamps) - 1]
        first_comment_timestamp = user_timestamps[0]

        difference = last_comment_timestamp - first_comment_timestamp
        L_d = int(float(difference) / (60 * 60 * 24))
        L_m = int(float(difference) / (60))
        L_w = int(float(difference) / (60 * 60 * 24 * 7))

        user_list = list(map(lambda t: int((t - first_comment_timestamp) / 60), user_timestamps))
        user_list.sort()

        user_list_map[users[i]] = user_list

        PDH, PWD = compute_time_measures(users[i])
        WS1, WS2, WS3 = compute_similarities(users[i])

        users[i].set_index(RegularityIndices.PDH, str(round(PDH, 2)))
        users[i].set_index(RegularityIndices.PWD, str(round(PWD, 2)))
        users[i].set_index(RegularityIndices.WS1, str(round(WS1, 2)))
        users[i].set_index(RegularityIndices.WS2, str(round(WS2, 2)))
        users[i].set_index(RegularityIndices.WS3, str(round(WS3, 3)))



from typing import List

from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.core.pos import POS as PosEum
from rb.cna.cna_graph import CnaGraph


# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.syntax.dep_enum import DepEnum
    from rb.complexity.morphology.pos_main import PosMain
    from rb.complexity.morphology.unq_pos_main import UnqPosMain
    from rb.complexity.morphology.pron_type_enum import PronounTypeEnum
    from rb.complexity.morphology.pron import Pronoun
    
    indices = []
    
    text_element_types = [TextElementType.SENT.value, TextElementType.BLOCK.value, TextElementType.DOC.value]
    measure_functions = [MeasureFunction.AVG, MeasureFunction.STDEV, MeasureFunction.MAX]
    parts_of_speech = [PosEum.NOUN, PosEum.VERB, PosEum.ADJ, PosEum.ADV, PosEum.PRON]
    
    for text_element in text_element_types:
        for pos in parts_of_speech:
            for measure_function in measure_functions:
                indices.append(PosMain(lang, pos, text_element, measure_function))
                indices.append(UnqPosMain(lang, pos, text_element, measure_function))
    
    if lang is Lang.RO or lang is Lang.EN or lang is Lang.RU:
        for pt in PronounTypeEnum:
            for text_element in text_element_types:
                for measure_function in measure_functions:
                    indices.append(Pronoun(lang, pt, text_element, measure_function))

    return indices


import tensorflow as tf
from rb.processings.gec.transformer.multi_head_attention import MultiHeadAttention
from rb.processings.gec.transformer.utils import point_wise_feed_forward_network

class EncoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1):
        super(EncoderLayer, self).__init__()

        self.mha = MultiHeadAttention(d_model, num_heads)
        self.ffn = point_wise_feed_forward_network(d_model, dff)

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)

    def call(self, x, training, mask):
        attn_output, _ = self.mha(x, x, x, mask)  # (batch_size, input_seq_len, d_model)
        attn_output = self.dropout1(attn_output, training=training)
        out1 = self.layernorm1(x + attn_output)  # (batch_size, input_seq_len, d_model)

        ffn_output = self.ffn(out1)  # (batch_size, input_seq_len, d_model)
        ffn_output = self.dropout2(ffn_output, training=training)
        out2 = self.layernorm2(out1 + ffn_output)  # (batch_size, input_seq_len, d_model)

        return out2


import os
from os import listdir
from os.path import isfile, join

import numpy as np
import tensorflow as tf
import tensorflow_datasets as tfds
from bert.tokenization.bert_tokenization import FullTokenizer
from google.cloud import storage

args = None

def _bytes_feature(value):
    """Returns a bytes_list from a string / byte."""
    if isinstance(value, type(tf.constant(0))):
        value = value.numpy() # BytesList won't unpack a string from an EagerTensor.
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

def _float_feature(value):
    """Returns a float_list from a float / double."""
    return tf.train.Feature(float_list=tf.train.FloatList(value=[value]))

def _int64_feature(value):
    """Returns an int64_list from a bool / enum / int / uint."""
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))

def _tensor_feature(value):
    """converts a tensor to serialized byte string"""
    return  _bytes_feature(tf.io.serialize_tensor(value))

def serialize_example(data, seg):
    """
    Creates a tf.Example message ready to be written to a file.
    """
    # Create a dictionary mapping the feature name to the tf.Example-compatible
    # data type.
    feature = {
        'data': _tensor_feature(data),
        'segs': _tensor_feature(seg),
    }

    # Create a Features message using tf.train.Example.
    example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
    return example_proto.SerializeToString()

def serialize_example_text(s, t):
    """
    Creates a tf.Example message ready to be written to a file.
    """
    # Create a dictionary mapping the feature name to the tf.Example-compatible
    # data type.
    feature = {
        'source': _bytes_feature(s),
        'target': _bytes_feature(t),
    }

    # Create a Features message using tf.train.Example.
    example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
    return example_proto.SerializeToString()

def serialize_example_ids(sentences, seg):
    """
    Creates a tf.Example message ready to be written to a file.
    """
    # Create a dictionary mapping the feature name to the tf.Example-compatible
    # data type.
    feature = {
        'sentences': _tensor_feature(sentences),
        'seg': _tensor_feature(seg),
    }

    # Create a Features message using tf.train.Example.
    example_proto = tf.train.Example(features=tf.train.Features(feature=feature))
    return example_proto.SerializeToString()

def tf_serialize_example_ids(sentences, seg):
    tf_string = tf.py_function(
        serialize_example_ids,
        (sentences, seg),  # pass these args to the above function.
        tf.string)      # the return type is `tf.string`.
    return tf.reshape(tf_string, ()) # The result is a scalar

def tf_serialize_example(data, seg):
    tf_string = tf.py_function(
        serialize_example,
        (data, seg),  # pass these args to the above function.
        tf.string)      # the return type is `tf.string`.
    return tf.reshape(tf_string, ()) # The result is a scalar

def example_encode_tensor():
    x = np.random.randint(0, high=128, size=(4, 2, 8))
    xx = np.random.randint(0, high=128, size=(4, 8))
    x = tf.convert_to_tensor(x, dtype=tf.int32)
    xx = tf.convert_to_tensor(xx, dtype=tf.int32)

    serialized_example = serialize_example(x, xx) # tensor ->  byte string -> tf.feature -> tf example -> serialized string
    example_proto = tf.train.Example.FromString(serialized_example) # serialized string -> tf example 
    print(example_proto)
    feature = example_proto.features.feature # tf example -> tf feature
    print(feature)
    segs = feature['segs'].bytes_list.value[0] # tf feature -> byte string (serialized 2) 
    print(segs)
    y = tf.io.parse_tensor(segs, out_type=tf.int32) # byte string -> tensor 
    print(y)
  
def example_encode_text():

    s = "asdqăđșßâ".encode('utf-8')
    t = "ăđßr€rțhy".encode('utf-8')

    serialized_example = serialize_example_text(s, t) # text ->  -> tf.feature -> tf example -> serialized string
    example_proto = tf.train.Example.FromString(serialized_example) # serialized string -> tf example 
    print(example_proto)
    feature = example_proto.features.feature # tf example -> tf feature
    print(feature)
    source = feature['source'].bytes_list.value[0] # tf feature -> text
    print(source.decode('utf-8'))

def generator_text():
    for _ in range(1024):
        s = 'asda'.encode('utf-8')
        t = 'asda'.encode('utf-8')
        yield serialize_example_text(s, t)

def parse_example(example):
    feature_description = {
        'source': tf.io.VarLenFeature(tf.string), # or tf.fixedLen
        'target': tf.io.VarLenFeature(tf.string)
    }
    y = tf.io.parse_single_example(example, feature_description) # get the tensor
    return (tf.sparse.to_dense(y['source'])[0], tf.sparse.to_dense(y['target'])[0])

def parse_example_ids(example):
    global args 
    # todo
    feature_description = {
        'sentences': tf.io.FixedLenFeature((), tf.string), # or tf.fixedLen
        'seg': tf.io.FixedLenFeature((), tf.string)
    }
    parsed_example = tf.io.parse_single_example(example, feature_description) # get the tensor

    seg = parsed_example['seg']
    seg = tf.io.parse_tensor(seg, out_type=tf.int64)
    seg = tf.reshape(seg, shape=(args.seq_length, ))

    sentences = parsed_example['sentences']
    sentences = tf.io.parse_tensor(sentences, out_type=tf.int64)
    sentences = tf.reshape(sentences, shape=(2, args.seq_length))
    return sentences, seg

def example_encode_text_dataset(args, filename='test.tfrecord'):
    serialized_features_dataset = tf.data.Dataset.from_generator(
        generator_text, output_types=tf.string, output_shapes=())
    # write the dataset to a file
    writer = tf.data.experimental.TFRecordWriter(filename)
    writer.write(serialized_features_dataset)
    # read from tf record file
    filenames = [filename]
    raw_dataset = tf.data.TFRecordDataset(filenames)
    # for x in raw_dataset.take(5):
    #     print(x)
    dataset = raw_dataset.map(parse_example)

        #print(x[0].numpy()[0].decode('utf-8'), x[1].numpy()[0].decode('utf-8'))
    return dataset

def serialize_ids_dataset(dataset, args, filename='train.tfrecord'):
    tf_records_path = os.path.join(args.tf_records, filename)
    serialized_dataset = dataset.map(tf_serialize_example_ids)
    writer = tf.data.experimental.TFRecordWriter(tf_records_path)
    writer.write(serialized_dataset)

def get_ids_dataset_tf_records(args1):
    global args
    args = args1
    # get dataset
    path_tf_records = args.tf_records
    if args.use_tpu:
        path_tf_records = join('gs://', args.bucket, path_tf_records)

    train_tf_record_file = join(path_tf_records, 'train.tfrecord')
    dev_tf_record_file = join(path_tf_records, 'dev.tfrecord')

    tf.compat.v1.logging.info('restoring tf records from {} {}'.format(train_tf_record_file, dev_tf_record_file))

    raw_train_dataset = tf.data.TFRecordDataset(train_tf_record_file)
    train_dataset = raw_train_dataset.map(parse_example_ids)

    raw_dev_dataset = tf.data.TFRecordDataset(dev_tf_record_file)
    dev_dataset = raw_dev_dataset.map(parse_example_ids)
    return train_dataset, dev_dataset

def get_tokenizers_ckeckpoint(args1):
    global args
    args = args1
    tokenizer_ro_path = join(args.checkpoint, 'tokenizer_ro')
    tokenizer_ro = tfds.features.text.SubwordTextEncoder.load_from_file(tokenizer_ro_path)
    tf.compat.v1.logging.info('restoring ro tokenizer from {}'.format(tokenizer_ro_path))

    tokenizer_bert = None
    if args.bert:
        tokenizer_bert_path = join(args.checkpoint, 'tokenizer_bert.vocab')
        tokenizer_bert = FullTokenizer(vocab_file=tokenizer_bert_path)
        tokenizer_bert.vocab_size = len(tokenizer_bert.vocab)
        tf.compat.v1.logging.info('restoring bert tokenizer from {}'.format(tokenizer_bert_path))

    tf.compat.v1.logging.info('tokenizers restored')
    return tokenizer_ro, tokenizer_bert 


def get_text_dataset_tf_records(path_tf_records):
    tf_records_files = [join(path_tf_records, f) for f in listdir(path_tf_records) \
            if isfile(join(path_tf_records, f)) and f.endswith('.tfrecord')]
    
    raw_dataset = tf.data.TFRecordDataset(tf_records_files)
    dataset = raw_dataset.map(parse_example)

    return dataset

def upload_blob(bucket_name, source_file_name, destination_blob_name):
    """Uploads a file to the bucket."""
    # bucket_name = "your-bucket-name"
    # source_file_name = "local/path/to/file"
    # destination_blob_name = "storage-object-name"

    storage_client = storage.Client()
    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(destination_blob_name)

    blob.upload_from_filename(source_file_name)
    tf.compat.v1.logging.info('file {} uploaded {} to bucket {}'.format(
        source_file_name, destination_blob_name, bucket_name))

if __name__ == "__main__":
   
    # example_encode_text()
    # example_encode_text_dataset(None)
    dataset = get_text_dataset_tf_records('./corpora/tf_records/test/')
    # for x, y in dataset:
    #     print(x)
    examples = [(s.numpy(), t.numpy()) for s, t in dataset]
    # for x, y in examples:
    #     print(x)
    # x = np.random.randint(0, high=128, size=(4, 2, 8))
    # xx = np.random.randint(high=128, size=(4, 8))

    # y = _tensor_feature(x)
    # print(y)
    # z = tf.train.BytesList(value=[y.numpy()])
    # print('bytes list:', tf.train.BytesList(value=[y.numpy()]))
    # print('tf train feature: ', tf.train.Feature(bytes_list=z))
    # x = tf.io.parse_tensor(y, out_type=tf.dtypes.int32)
    # print(_bytes_feature(b'test_string'))
    # print(_bytes_feature(u'test_bytes'.encode('utf-8')))



from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoTenseEnum(Enum):
    FUT = auto()
    IMP = auto()
    PAST = auto()
    PQP = auto()
    PRES = auto()


class RoPOSFeatureTense(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.TENSE,
                            values=None, pos_supported=None)
        self.values = set([ro_tense for _, ro_tense in RoTenseEnum.__members__.items()])
        self.pos_supported = set([POS.PART, POS.AUX, POS.VERB, POS.X])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureTense":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureTense()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoTenseEnum]:
        pattern = '(Tense=[a-zA-Z0-9,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nrs = matches[0].split('=')[1].split(',')
        values = []
        for nr in nrs:
            if nr == 'Fut':
                values.append(RoTenseEnum.FUT)
            elif nr == 'Imp':
                values.append(RoTenseEnum.IMP)
            elif nr == 'Past':
                values.append(RoTenseEnum.PAST)
            elif nr == 'Pqp':
                values.append(RoTenseEnum.PQP)
            elif nr == 'Pres':
                values.append(RoTenseEnum.PRES)
        return values


from enum import Enum
from typing import List

from rb.complexity.morphology.factory import create as morphology
from rb.complexity.surface.factory import create as surface
from rb.complexity.syntax.factory import create as syntax
from rb.complexity.word.factory import create as word
from rb.complexity.cohesion.factory import create as cohesion
from rb.complexity.rhythm.factory import create as rhythm
from rb.complexity.discourse.factory import create as discourse
from rb.cna.cna_graph import CnaGraph
from rb.core.lang import Lang


class IndexCategory(Enum):
    SURFACE = 0
    MORPHOLOGY = 1
    SYNTAX = 2
    WORD = 3
    COHESION = 4
    RHYTHM = 5
    DISCOURSE = 6

    def create(self, lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
        functions = [surface, morphology, syntax, word, cohesion, rhythm, discourse]
        return functions[self.value](lang, cna_graph)



from enum import Enum, auto

class RoFeaturesName(Enum):
    ABBR = 'Abbreviation'
    ADPT = 'AdpType'
    CASE = 'Case'
    DEF = 'Definite'
    DEG = 'Degree'
    FGN = 'Foreign'
    GEN = 'Gender'
    MOOD = 'Mood'
    NR = 'Number'
    NRP = 'Number[psor]'
    NRF = 'NumForm'
    NRT = 'NumType'
    PARTT = 'PartType'
    PER = 'Person'
    POL = 'Polarity'
    POSI = 'Position'
    POSS = 'Poss'
    PRONT = 'PronType'
    RFL = 'Reflex'
    STRN = 'Strength'
    TENSE = 'Tense'
    VAR = 'Variant'
    VBF = 'VerbForm'
     


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoNumberEnum(Enum):
    PLUR = auto()
    SING = auto()


class RoPOSFeatureNumber(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.NR,
                            values=None, pos_supported=None)
        self.values = set([ro_nr for _, ro_nr in RoNumberEnum.__members__.items()])
        self.pos_supported = set([POS.ADJ, POS.AUX, POS.DET, POS.NOUN, POS.NUM, POS.PRON, POS.PROPN, POS.VERB])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureNumber":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureNumber()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoNumberEnum]:
        pattern = '(Number=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nrs = matches[0].split('=')[1].split(',')
        values = []
        for nr in nrs:
            if nr == 'Plur':
                values.append(RoNumberEnum.PLUR)
            elif nr == 'Sing':
                values.append(RoNumberEnum.SING)
        return values


import networkx as nx
import pandas as pd
from tabulate import tabulate
from sklearn.cluster import AgglomerativeClustering
import matplotlib.pyplot as plt
from pprint import pprint
import numpy as np


class GraphMetrics:

    def __init__(self, graph: "Graph"):
        self.graph = graph
        self.semantic_distances_between_articles = graph.get_distances_between_articles()
        self.semantic_distances_between_authors = graph.get_distances_between_authors()
        # self.semantic_similarities_between_articles = graph.get_similarities_between_articles()
        # self.semantic_similarities_between_authors = graph.get_similarities_between_authors()

    def compute_articles_closeness(self):
        distances_graph = build_distance_graph(self.semantic_distances_between_articles)
        return dict(nx.closeness_centrality(distances_graph, distance='weight'))

    def compute_articles_betweenness(self):
        distances_graph = build_distance_graph(self.semantic_distances_between_articles)
        return dict(nx.betweenness_centrality(distances_graph, weight='weight', normalized=True))

    def compute_authors_closeness(self, graph_type='bidirectional'):
        distances_graph = build_distance_graph(self.semantic_distances_between_authors, graph_type)
        return dict(nx.closeness_centrality(distances_graph, distance='weight'))

    def compute_authors_betweenness(self, graph_type='bidirectional'):
        distances_graph = build_distance_graph(self.semantic_distances_between_authors, graph_type)
        return dict(nx.betweenness_centrality(distances_graph, weight='weight', normalized=True))

    def get_top_n_articles_by_closeness(self, n):
        closeness = self.compute_articles_closeness()
        articles_closeness = [(k, v) for k, v in sorted(closeness.items(), key=lambda item: item[1], reverse=True)][:n]
        return articles_closeness

    def get_top_n_articles_by_betweenness(self, n):
        betweenness = self.compute_articles_betweenness()
        articles_betweenness = [(k, v)
                                for k, v in sorted(betweenness.items(), key=lambda item: item[1], reverse=True)][:n]
        return articles_betweenness

    def get_top_n_authors_by_closeness(self, n, graph_type='bidirectional'):
        closeness = self.compute_authors_closeness(graph_type)
        authors_closeness = [(k, v) for k, v in sorted(closeness.items(), key=lambda item: item[1], reverse=True)][:n]
        return authors_closeness

    def get_top_n_authors_by_betweenness(self, n, graph_type='bidirectional'):
        betweenness = self.compute_authors_betweenness(graph_type)
        authors_betweenness = [(k, v)
                               for k, v in sorted(betweenness.items(), key=lambda item: item[1], reverse=True)][:n]
        return authors_betweenness

    def perform_articles_agglomerative_clustering(self):
        distance_threshold = self.graph.articles_mean
        return perform_agglomerative_clustering(self.graph.articles_set, self.semantic_distances_between_articles,
                                                distance_threshold)

    def perform_authors_agglomerative_clustering(self):
        distance_threshold = self.graph.authors_mean - self.graph.authors_std
        print(self.graph.authors_mean, self.graph.authors_std)
        return perform_agglomerative_clustering(self.graph.authors_set, self.semantic_distances_between_authors,
                                                distance_threshold)

    def get_top_clusters_of_articles_by_number_of_elements(self, number_of_clusters):
        return self.get_top_clusters_of_articles_by_criteria(number_of_clusters, sort_clusters_by_size, self.graph)

    def get_top_clusters_of_articles_by_degree_of_elements(self, number_of_clusters):
        return self.get_top_clusters_of_articles_by_criteria(number_of_clusters, sort_clusters_by_degree, self.graph)

    def get_top_clusters_of_articles_by_criteria(self, number_of_clusters, criteria_function, graph):
        clusters = self.perform_articles_agglomerative_clustering()
        sorted_clusters = criteria_function(clusters, graph.get_articles_by_type_degree)
        counter = 0
        results = []
        for label, articles in sorted_clusters.items():
            titles = ",\n".join([article.title for article in articles])
            abstracts = "\n".join([article.abstract for article in articles])
            results.append((titles, abstracts))
            counter += 1
            if counter >= number_of_clusters:
                break
        return results


def build_distance_graph(distances_pairs, graph_type='bidirectional'):
    if graph_type == 'bidirectional':
        distance_graph = nx.Graph()
    else:
        distance_graph = nx.DiGraph()
    for pair in distances_pairs:
        entity1 = pair[0]
        entity2 = pair[1]
        distance = pair[2]
        distance_graph.add_node(entity1)
        distance_graph.add_node(entity2)
        distance_graph.add_edge(entity1, entity2, weight=distance)
    return distance_graph


def build_position_dictionary(entity_set):
    index = 0
    position_dictionary = {}
    for entity in entity_set:
        position_dictionary[entity] = index
        index += 1
    return position_dictionary


def build_distance_matrix(distances_pairs, side, position_dictionary):
    distance_matrix = [[1 for _ in range(side)] for _ in range(side)]
    for pair in distances_pairs:
        row = position_dictionary[pair[0]]
        column = position_dictionary[pair[1]]
        distance_matrix[row][column] = pair[2]
    return distance_matrix


def build_inverse_dictionary(position_dictionary):
    inverse_dictionary = {}
    for key, value in position_dictionary.items():
        inverse_dictionary[value] = key
    return inverse_dictionary


def perform_agglomerative_clustering(entity_set, semantic_distances, distance_threshold):
    position_dictionary = build_position_dictionary(entity_set)
    side = len(entity_set)
    distance_matrix = build_distance_matrix(semantic_distances, side, position_dictionary)
    model = AgglomerativeClustering(affinity='precomputed', n_clusters=None,
                                    distance_threshold=distance_threshold, linkage="single")
    results = model.fit(distance_matrix)
    print(max(results.labels_))
    # plot_clustering_labels(results.labels_, side)
    return structure_agglomerative_clustering_results(results, build_inverse_dictionary(position_dictionary))


def structure_agglomerative_clustering_results(results, inverse_position_dictionary):
    clusters = {}
    for index, label in enumerate(results.labels_):
        clusters[label] = clusters.get(label, [])
        clusters[label].append(inverse_position_dictionary[index])  # add name or title if you want that
    # pprint(clusters)
    return clusters


def plot_clustering_labels(labels, side):
    plt.scatter(range(side), range(side), c=labels, cmap='rainbow')
    plt.show()


def build_degree_dictionary(rankings):
    degree_dictionary = {}
    for info in rankings:
        degree_dictionary[info[0]] = info[1]
    return degree_dictionary


def sort_clusters_by_size(clusters, not_used):
    return {k: v for k, v in sorted(clusters.items(), key=lambda item: len(item[1]), reverse=True)}


def sort_clusters_by_degree(clusters, get_rankings_function):
    rankings = get_rankings_function()
    degree_dictionary = build_degree_dictionary(rankings)
    return {k: v for k, v in sorted(clusters.items(), key=lambda item: np.mean([degree_dictionary.get(node, 0)
                                                                               for node in item[1]]), reverse=True)}



import re

RE_WHITESPACE = re.compile(r"(\s)+", re.UNICODE)


def merge_and_rewrite():
    path1 = "/home/vacioaca/Downloads/s50_0_words_all.txt"
    path2 = "/home/vacioaca/Downloads/s50_1_words_all.txt"
    dest_path = "/home/vacioaca/Downloads/syllabified_dict.dict"
    dictionary = {}
    with open(path1, mode="r", encoding="utf-8-sig") as file1:
        for line in file1.readlines():
            line = RE_WHITESPACE.sub(" ", line.strip())
            if len(line) == 0:
                continue
            pieces = line.split(" ", 1)
            word = pieces[1].split(" ", 1)[0]
            syllables = pieces[0].split("-")
            if word not in dictionary:
                dictionary[word] = [syllables]
            else:
                filtered_syllables = filter(lambda t: set(t) == set(syllables), dictionary[word])
                if next(filtered_syllables, None) is None:
                    dictionary[word] += [syllables]

    with open(path2, mode="r", encoding="utf-8-sig") as file2:
        for line in file2.readlines():
            line = RE_WHITESPACE.sub(" ", line.strip())
            if len(line) == 0:
                continue
            pieces = line.split(" ", 1)
            word = pieces[1].split(" ", 1)[0]
            syllables = pieces[0].split("-")
            if word not in dictionary:
                dictionary[word] = [syllables]
            else:
                filtered_syllables = filter(lambda t: set(t) == set(syllables), dictionary[word])
                if next(filtered_syllables, None) is None:
                    dictionary[word] += [syllables]

    with open(dest_path, mode="w", encoding="utf-8") as file:
        for key in sorted(dictionary.keys()):
            for syllables in dictionary[key]:
                file.write("{} {}\n".format(key, "-".join(syllables)))



import tensorflow as tf
gpus = tf.config.experimental.list_physical_devices('GPU')
if gpus:
    try:
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)

    except RuntimeError as e:
        print(e)
import tensorflow.keras as keras
import numpy as np
import BertModel
import utils
import sys
import absl
import bert
import functools
import os


FLAGS = absl.flags.FLAGS

absl.flags.DEFINE_integer('max_seq_len', 128, 'Maximum sequence length')
absl.flags.DEFINE_string('model_folder_path', '../Models/raw/multi_cased_base/', 'Path to bert model folder')
absl.flags.DEFINE_string('model_weights_path', 'models/finetune_sentiment_emag/multi_cased_base/2/model.h5', "Path to bert finetuned weights")
absl.flags.DEFINE_integer('dev_batch_size', 32, 'Batch size to use during evaluation on dev')
absl.flags.DEFINE_integer('test_batch_size', 32, 'Batch size to use during evaluation on test')
absl.flags.DEFINE_integer('num_classes', 4, "Number of classes for clasification task")
absl.flags.DEFINE_bool("use_tpu", False, "Use TPU or not")
absl.flags.DEFINE_string("tpu_name", None, "Name of TPU instance")


def create_model():
    # define input
    input_ids = tf.keras.layers.Input(shape=(FLAGS.max_seq_len), dtype=tf.int32, name="input_ids")
    # no need for attention mask; it is computed in automatically in bert model
    # attention_mask = tf.keras.layers.Input(shape=(FLAGS.max_seq_len), dtype=tf.int32, name="input_ids")
    segment_ids = tf.keras.layers.Input(shape=(FLAGS.max_seq_len), dtype=tf.int32, name="segment_ids")

    do_lower_case = True
    if "multi_cased_base" in FLAGS.model_folder_path:
        do_lower_case = False

    # define model
    bert_model = BertModel.BertModel(FLAGS.model_folder_path, FLAGS.max_seq_len, do_lower_case)
    bert_output = bert_model.bert_layer([input_ids, segment_ids])
    cls_output = keras.layers.Lambda(lambda seq: seq[:, 0, :])(bert_output)
    fc1 = keras.layers.Dense(units=100, activation="relu")(cls_output)
    prediction = keras.layers.Dense(units=FLAGS.num_classes, activation="softmax")(fc1)

    # build model
    model = keras.Model(inputs=[input_ids, segment_ids], outputs=prediction)
    model.build(input_shape=[(None, FLAGS.max_seq_len), (None, FLAGS.max_seq_len)])
    # load pretrained
    model.load_weights(FLAGS.model_weights_path)
    
    model.compile(loss = keras.losses.CategoricalCrossentropy(from_logits=False), metrics = [tf.keras.metrics.categorical_accuracy])
    model.summary()

    print("Do lower case =", do_lower_case)


    return model, bert_model


def create_tfdataset(features, labels, split, batch_size):

    aux1 = []
    aux2 = []
    for i in range(len(features[0])):
        aux1.append(features[0][i])
        aux2.append(features[1][i])

    dataset = tf.data.Dataset.from_tensor_slices(((aux1, aux2), labels))
    if split == "train":
        dataset = dataset.shuffle(22000)
        dataset = dataset.repeat(-1).batch(batch_size)

    else:
        dataset = dataset.batch(batch_size)

    return dataset

def main(argv):
    del argv
    
    # create model
    if FLAGS.use_tpu == True:
        tpu_cluster_resolver = tf.distribute.cluster_resolver.TPUClusterResolver(FLAGS.tpu_name, zone=None, project=None)
        tf.config.experimental_connect_to_cluster(tpu_cluster_resolver)
        tf.tpu.experimental.initialize_tpu_system(tpu_cluster_resolver)
        strategy = tf.distribute.experimental.TPUStrategy(tpu_cluster_resolver)
        with strategy.scope():
            model, bert = create_model()
    else:
            model, bert = create_model()
    
    # load data
    dev_data = utils.readJson("../Dataset/Reviews/4Classes/dev.json")
    test_data = utils.readJson("../Dataset/Reviews/4Classes/test.json")

    dev_features = utils.getFeatures(dev_data)
    dev_features, dev_labels, _ = utils.processFeatures(dev_features, bert)
    print(len(dev_features[0]), len(dev_labels))

    test_features = utils.getFeatures(test_data)
    test_features, test_labels, _ = utils.processFeatures(test_features, bert)
    print(len(test_features[0]), len(test_labels))
    
    dev_dataset = create_tfdataset(dev_features, dev_labels, "dev", FLAGS.dev_batch_size)
    test_dataset = create_tfdataset(test_features, test_labels, "test", FLAGS.test_batch_size)

    fmetric_dev = utils.FScoreCallback(dev_dataset, len(dev_labels)//FLAGS.dev_batch_size, dev_labels)
    model.evaluate(dev_dataset, callbacks=[fmetric_dev])

    fmetric_test = utils.FScoreCallback(test_dataset, len(test_labels)//FLAGS.test_batch_size, test_labels)
    model.evaluate(test_dataset, callbacks=[fmetric_test])

if __name__ == "__main__":

    absl.flags.mark_flag_as_required('model_folder_path')
    absl.app.run(main)



from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.similarity.wordnet import get_all_paths_lengths_to_root
from typing import List, Callable
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class WdMaxDpthHypTree(ComplexityIndex):
    
    def __init__(self, lang: Lang, reduce_depth: int,
            reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.WORD,
                                 abbr="MaxDepthHypTree", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

    def _compute_value(self, element: TextElement) -> int:
        if element.is_content_word():
            paths = get_all_paths_lengths_to_root(element)
            return max(paths) if paths else 0
        else:
            return 0


from rb.processings.pipeline.estimator import Classifier
from rb.processings.pipeline.dataset import Dataset, Task
from typing import List, Dict
from sklearn.ensemble import RandomForestClassifier

class RandomForest(Classifier):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.model = RandomForestClassifier(n_estimators=params["n_estimators"])
        self.n_estimators: int = params["n_estimators"]
    
    @classmethod
    def parameters(cls):
        return {
            "n_estimators": [32, 64, 96, 128]
        }
        
    def __str__(self):
        return f"Random forest - {self.n_estimators} estimators"


from rb.core.sentence import Sentence
from rb.core.word import Word
from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO
from rb.comprehension.utils.graph.cm_node_type import CmNodeType
from rb.comprehension.utils.graph.cm_edge_do import CmEdgeDO, CmEdgeType

from typing import List, Dict, Set

class CmDependencyGraf:

    def __init__(self):
        self.node_list: Set[Word] = set([])
        self.adjacent_list: Dict[Word, Set[Word]] = {}

    def get_actual_word(self, word):
        if word.in_coref:
            return word.coref_clusters[0].main.get_root()
        return word

    def build_graph(self, sentence: Sentence) -> None:
        for index, dependency in enumerate(sentence.get_dependencies()):
            w0 = self.get_actual_word(dependency[0])
            w1 = self.get_actual_word(dependency[1])
            self.node_list.add(w0)
            self.node_list.add(w1)
            if w0 in self.adjacent_list:
                self.adjacent_list[w0].add(w1)
            else:
                self.adjacent_list[w0] = set([w1])

            if w1 in self.adjacent_list:
                self.adjacent_list[w1].add(w0)
            else:
                self.adjacent_list[w1] = set([w0])

    def dfs(self, node: Word) -> List[Word]:
        results = []
        stack = []
        visited = set([])
        
        stack.append(node)
        visited.add(node)
        while stack:
            word = stack.pop()
            for neighbour in self.adjacent_list[word]:
                if neighbour not in visited:
                    if neighbour.is_content_word():
                        results.append(neighbour)
                    else:
                        stack.append(neighbour)
                    visited.add(neighbour)

        return results

    def get_syntactic_graph(self, sentence: Sentence) -> CmGraphDO:
        self.build_graph(sentence)
        graph = CmGraphDO([], [])
        node_set = set([])
        for node in self.node_list:
            if node.is_content_word():
                accessible_content_words = self.dfs(node)
                n1 = CmNodeDO(node, CmNodeType.TextBased)
                n1.activate()
                node_set.add(n1)
                for word in accessible_content_words:
                    n2 = CmNodeDO(word, CmNodeType.TextBased)
                    n2.activate()
                    node_set.add(n2)

                    edge = CmEdgeDO(n1, n2, CmEdgeType.Syntactic, 1)

                    if edge not in graph.edge_list:
                        graph.edge_list.append(edge)
        
        graph.node_list = list(node_set)

        return graph



from rb.complexity.complexity_index import ComplexityIndex
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.complexity.rhythm.syllabified_dict import SyllabifiedDict
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.utils.rblogger import Logger

from typing import List

logger = Logger.get_logger()


class NoSyllablesIndex(ComplexityIndex):

    def __init__(self, lang: Lang, reduce_depth: int,
                 reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.RHYTHM,
                                 abbr="RhNoSyll", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

        self.syllabified_dict = SyllabifiedDict.get_instance(lang)

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> float:
        if element.is_word():
            if element.text in self.syllabified_dict:
                # a word can be syllabified in multiple ways
                syllabified_words = self.syllabified_dict[element.text]
                if len(syllabified_words) == 1:
                    syllables = syllabified_words[0]
                    return len(syllables)
                else:
                    # TODO: choose the correct version
                    pass
            else:
                return 1
        elif element.depth <= self.reduce_depth:
            res = 0
            for child in element.components:
                res += self.compute_below(child)
            return res

    def compute_above(self, element: TextElement) -> List[float]:
        values = list()
        if element.depth > self.reduce_depth:
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = [self.compute_below(element)]
        else:
            logger.error("wrong reduce depth value.")
        return values




class WordActivation():

    def __init__(self, activation_value: float, active: bool):
        self.activation_value = activation_value
        self.active = active

    def __repr__(self):
        return "WordActivation(%r, active: %r)" % (self.activation_value, self.active)


    def __str__(self):
        return "WordActivation(%r, active: %r)" % (self.activation_value, self.active)


import tensorflow as tf
from rb.processings.gec.transformer.multi_head_attention import MultiHeadAttention
from rb.processings.gec.transformer.utils import point_wise_feed_forward_network

class DecoderLayer(tf.keras.layers.Layer):
    def __init__(self, d_model, num_heads, dff, rate=0.1):
        super(DecoderLayer, self).__init__()

        self.mha1 = MultiHeadAttention(d_model, num_heads)
        self.mha2 = MultiHeadAttention(d_model, num_heads)

        self.ffn = point_wise_feed_forward_network(d_model, dff)

        self.layernorm1 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm2 = tf.keras.layers.LayerNormalization(epsilon=1e-6)
        self.layernorm3 = tf.keras.layers.LayerNormalization(epsilon=1e-6)

        self.dropout1 = tf.keras.layers.Dropout(rate)
        self.dropout2 = tf.keras.layers.Dropout(rate)
        self.dropout3 = tf.keras.layers.Dropout(rate)


    def call(self, x, enc_output, training, 
            look_ahead_mask, padding_mask):
        # enc_output.shape == (batch_size, input_seq_len, d_model)

        attn1, attn_weights_block1 = self.mha1(x, x, x, look_ahead_mask)  # (batch_size, target_seq_len, d_model)
        attn1 = self.dropout1(attn1, training=training)
        out1 = self.layernorm1(attn1 + x)

        attn2, attn_weights_block2 = self.mha2(
            enc_output, enc_output, out1, padding_mask)  # (batch_size, target_seq_len, d_model)
        attn2 = self.dropout2(attn2, training=training)
        out2 = self.layernorm2(attn2 + out1)  # (batch_size, target_seq_len, d_model)

        ffn_output = self.ffn(out2)  # (batch_size, target_seq_len, d_model)
        ffn_output = self.dropout3(ffn_output, training=training)
        out3 = self.layernorm3(ffn_output + out2)  # (batch_size, target_seq_len, d_model)

        return out3, attn_weights_block1, attn_weights_block2


from typing import Dict, List, Tuple
from datetime import datetime, timedelta

from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.cscl.participant import Participant
from rb.similarity.vector_model_factory import create_vector_model
from rb.similarity.vector_model import VectorModelType
from rb.utils.rblogger import Logger
from copy import deepcopy

logger = Logger.get_logger()

# muchii intre replici parinte-copil?
# muchii intre replici ale aceluiasi user ?

class Community(TextElement):

    '''
        community -> list of JSON objects (crawled from Reddit), with
                     the same structure as described in Conversation class
    '''
    def __init__(self, lang: Lang, 
                 community: List[Dict],
                 container: TextElement = None,
                 depth: int = TextElementType.COMM.value,
                 start_date: int = None,
                 end_date: int = None):

        TextElement.__init__(self, lang=lang, text=None,
                             depth=depth, container=container)

        self.components = []
        self.start_date = start_date
        self.end_date = end_date

        self.participant_map = dict()
        self.conversation_map = {}
        self.community_raw = community
        
        
        for conversation_thread in deepcopy(community):
            current_conversation = Conversation(lang, container=self,
                                                 conversation_thread=conversation_thread,
                                                 start_date=start_date, end_date=end_date)
            if current_conversation.get_contributions():
                self.components.append(current_conversation)
                self.conversation_map[current_conversation.conv_id] = current_conversation

        self.participants = self.union_participants()
        self.participant_contributions = self.union_contributions()
        self.first_contribution_date, self.last_contribution_date = self.find_contribution_range()

        self.graph = None

        self.eligible_contributions = []
        self.timeframe_subcommunities = []
        self.scores = dict()
        self.init_scores()
        
    def union_participants(self) -> List[Participant]:
        return [self.participant_map[participant_id] for participant_id in self.participant_map]

    def union_contributions(self) -> Dict:
        contributions = dict()

        for conversation in self.components:
            for participant in conversation.get_participants():
                if participant.get_id() not in contributions:
                    contributions[participant.get_id()] = []

                contributions[participant.get_id()] += conversation.get_participant_contributions(participant.get_id())

        return contributions

    def get_community_raw(self) -> List[Dict]:
        return self.community_raw

    def find_contribution_range(self) -> Tuple[int, int]:
        first_contribution = None
        last_contribution = None

        for conversation in self.components:
            for contribution in conversation.get_contributions():
                timestamp = contribution.get_timestamp()

                if self.is_eligible(timestamp):
                    if first_contribution == None or timestamp < first_contribution:
                        first_contribution = timestamp

                    if last_contribution == None or timestamp > last_contribution:
                        last_contribution = timestamp

        if self.start_date is None:
            self.start_date = first_contribution
        if self.end_date is None:
            self.end_date = last_contribution

        return first_contribution, last_contribution

    def is_eligible(self, timestamp: int) -> bool:
        if self.start_date != None and self.end_date != None:
            return (timestamp >= self.start_date and timestamp <= self.end_date)
        
        return True

    def add_eligible_contribution(self, contribution: Contribution):
        self.eligible_contributions.append(contribution)

    def add_subcommunity(self, community):
        self.timeframe_subcommunities.append(community)

    def get_conversations(self) -> List[Conversation]:
        return self.components

    def get_conversation(self, conv_id) -> Conversation:
        return self.conversation_map[conv_id]
    
    def get_participant(self, participant_id: str) -> Participant:
        return self.participant_map[participant_id]

    def get_participants(self) -> List[Participant]:
        return self.participants

    def get_participant_contributions(self, participant_id: str) -> List[Contribution]:
        return self.participant_contributions[participant_id]

    def init_scores(self):
        for a in self.participants:
            self.scores[a.get_id()] = dict()

            for b in self.participants:
                self.scores[a.get_id()][b.get_id()] = 0

    def get_score(self, a: str, b: str) -> float:
        return self.scores[a][b]

    def update_score(self, a: str, b: str, value: float):
        self.scores[a][b] += value



import unittest
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.pos import POS

class CoreTest(unittest.TestCase):

    def test_create_document_en(self):
        text_string = "This is a text string. Does the parsing work?"
        doc = Document(Lang.EN, text_string)
        self.assertEqual(doc.text, text_string, "Should be " + text_string)

    def test_create_document_ro(self):
        text_string = "Acesta este un text de test. Întrebarea este: merge?"
        doc = Document(Lang.RO, text_string)
        self.assertEqual(doc.text, text_string, "Should be " + text_string)

    def test_get_words(self):
        text_string = "This is a text string. Does the parsing work?\nSecond paragraph."
        doc = Document(Lang.EN, text_string)
        self.assertEqual(len(doc.get_words()), 14, "Should be 14: " + str(doc.get_words()))

    def test_coref_en(self):
        text_string = "John is a programmer. He earns big money."
        doc = Document(Lang.EN, text_string)
        for word in doc.get_words():
            if word.text != "He":
                continue
            else:
                self.assertEqual(True, word.in_coref, "Should be: True")
                self.assertEqual("John", word.coref_clusters[0].main.get_root().text, "Should be: John")
        

if __name__ == '__main__':
    unittest.main()


import bisect
import math
from collections import Counter
from typing import Dict, List, Set, Tuple, Union

from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.text_element_type import TextElementType
from rb.core.word import Word
from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.similarity.vector_model_factory import create_vector_model, get_default_model
from rb.utils.rblogger import Logger
from rb.core.text_element import TextElement

logger = Logger.get_logger()


def extract_keywords(text: Union[str, TextElement], lang: Lang = Lang.RO, max_keywords: int = 40, vector_model: VectorModel = None, threshold: float = 0.3) -> List[Tuple[float, str]]:

    if vector_model is None:
        vector_model = get_default_model(lang)
    logger.info('Computing keywords...')
    if not isinstance(text, TextElement):
        doc = Document(lang=lang, text=text)
    else:
        doc = text
    lemma_words, raw_words = [], []
    scores: List[Tuple[float, Word]] = []

    for word in doc.get_words():
        if word.is_content_word() and not word.is_stop:
            raw_words.append(word)

    counter = Counter(raw_words)
    
    score_dict = {}
    for word, freq in counter.items():
        if word.lemma not in score_dict:
            score_dict[word.lemma] = 0
        freq = 1 + math.log(freq)
        sim = vector_model.similarity(word, doc)
        score_dict[word.lemma] += sim * freq

    scores = [(score, lemma) for lemma, score in score_dict.items()]
    scores = sorted(scores, reverse=True, key=lambda x: x[0])
    scores.reverse()
    idx = bisect.bisect(scores, (threshold, ))
    scores.reverse()
    if max_keywords == -1:
        return scores
    return scores[:len(scores)-idx]
    # return scores[:max_keywords]




from enum import Enum
from functools import partial
from statistics import mean, stdev
from typing import List

def average(elements: List[float]) -> float:
    elements = [elem for elem in elements if elem is not None]
    if len(elements) == 0:
        return None
    return mean(elements)

def standard_deviation(elements: List[float]) -> float:
    elements = [elem for elem in elements if elem is not None]
    if len(elements) == 0:
        return None
    elif len(elements) == 1:
        return 0
    return stdev(elements)

def maximum(elements: List[float]) -> float:
    elements = [elem for elem in elements if elem is not None]
    if len(elements) == 0:
        return None
    return max(elements)

def identity(elements: List[float]) -> float:
    return elements[0]

class MeasureFunction(Enum):
    AVG = average
    STDEV = standard_deviation
    MAX = maximum
    IDENTITY = identity


from itertools import combinations
import networkx as nx
import pandas as pd
from tabulate import tabulate


class CoAuthorship:

    def __init__(self, graph: "Graph"):
        self.co_authorship_dict = {}
        for article in graph.articles_set:
            self.create_co_authorship_for_article_authors(article, graph)

        self.NO_OF_AUTHORS_TO_PRINT = 1000

    def create_co_authorship_for_article_authors(self, article, graph):
        for (author1_name, author2_name) in list(combinations(article.author_list, r=2)):
            self.update_authors_co_authorship_count(author1_name, author2_name, graph)

    def update_authors_co_authorship_count(self, author1_name, author2_name, graph):
        author1 = graph.authors_dict[author1_name]
        author2 = graph.authors_dict[author2_name]
        common_articles_number = self.co_authorship_dict.get((author1, author2), 0)
        self.co_authorship_dict[(author1, author2)] = common_articles_number + 1

    def build_co_authorship_degree_graph(self):
        co_authorship_degree_graph = nx.Graph()
        for (author1, author2), co_authorship_distance in self.co_authorship_dict.items():
            co_authorship_degree_graph.add_node(author1)
            co_authorship_degree_graph.add_node(author2)
            co_authorship_degree_graph.add_edge(author1, author2, weight=co_authorship_distance)
        return co_authorship_degree_graph

    def get_rankings(self):
        co_authorship_degree_graph = self.build_co_authorship_degree_graph()
        co_authorship_degrees = dict(nx.degree(co_authorship_degree_graph, weight='weight'))
        return build_rankings_dictionary(co_authorship_degrees)

    def print_results(self):
        rankings_dictionary = self.get_rankings()
        co_authorship_degrees_df = pd.DataFrame.from_dict(rankings_dictionary)
        sorted_df = co_authorship_degrees_df.sort_values(by=['Degree'], ascending=False).head(
            self.NO_OF_AUTHORS_TO_PRINT)[["AuthorName", "NoOfArticles", "Degree"]]
        print(tabulate(sorted_df, headers='keys', tablefmt='psql', showindex=False))


def build_rankings_dictionary(co_authorship_degrees):
    rankings_dictionary = {"AuthorName": [], "NoOfArticles": [], "Degree": []}
    for author, degree in co_authorship_degrees.items():
        rankings_dictionary["AuthorName"].append(author.name)
        rankings_dictionary["NoOfArticles"].append(len(author.articles))
        rankings_dictionary["Degree"].append(degree)
    return rankings_dictionary




""" do not move this code in vector_model.py, otherwise you will have circular imports"""
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.similarity.lda import LDA
from rb.similarity.lsa import LSA
from rb.similarity.transformers_encoder import TransformersEncoder
from rb.similarity.vector import Vector
from rb.similarity.vector_model import (CorporaEnum, VectorModel,
                                        VectorModelType)
from rb.similarity.word2vec import Word2Vec
from rb.utils.rblogger import Logger

logger = Logger.get_logger() 


VECTOR_MODELS = {
    Lang.RO: {
        CorporaEnum.README: {
            VectorModelType.WORD2VEC: Word2Vec
        }
    },
    Lang.EN: {
        CorporaEnum.COCA: {
            VectorModelType.WORD2VEC: Word2Vec,
            VectorModelType.LDA: LDA,
            VectorModelType.LSA: LSA
        }
    },
    Lang.ES: {
        CorporaEnum.JOSE_ANTONIO: {
            VectorModelType.WORD2VEC: Word2Vec
        }
    },
    Lang.RU: {
        CorporaEnum.RNC_WIKIPEDIA: {
            VectorModelType.WORD2VEC: Word2Vec
        }
    },
    Lang.FR: {
        CorporaEnum.LE_MONDE: {
            VectorModelType.WORD2VEC: Word2Vec
        }
    },
    Lang.NL: {
        CorporaEnum.WIKI: {
            VectorModelType.WORD2VEC: Word2Vec
        }
    },
    Lang.DE: {
        CorporaEnum.WIKI: {
            VectorModelType.WORD2VEC: Word2Vec
        }
    },
}

DEFAULT_MODELS = {
    Lang.RO: (VectorModelType.WORD2VEC, CorporaEnum.README),
    Lang.EN: (VectorModelType.WORD2VEC, CorporaEnum.COCA),
    Lang.ES: (VectorModelType.WORD2VEC, CorporaEnum.JOSE_ANTONIO),
    Lang.RU: (VectorModelType.WORD2VEC, CorporaEnum.RNC_WIKIPEDIA),
    Lang.FR: (VectorModelType.WORD2VEC, CorporaEnum.LE_MONDE),
    Lang.NL: (VectorModelType.WORD2VEC, CorporaEnum.WIKI),
    Lang.DE: (VectorModelType.WORD2VEC, CorporaEnum.WIKI),
}

EXISTING_MODELS = {}


def save_model(vector_model: VectorModel, lang: Lang, model_type: VectorModelType, 
                corpus: str, dim: int, model_abbr: str) -> bool:
    global EXISTING_MODELS, logger

    if lang not in EXISTING_MODELS:
        EXISTING_MODELS[lang] = {}
    if model_type not in EXISTING_MODELS[lang]:
        EXISTING_MODELS[lang][model_type] = {}
    if corpus not in EXISTING_MODELS[lang][model_type]:
        EXISTING_MODELS[lang][model_type][corpus] = {}
    if dim not in EXISTING_MODELS[lang][model_type][corpus]:
        logger.info(f'Model {model_abbr} does not exists in memory. Saving it...')
        EXISTING_MODELS[lang][model_type][corpus][dim] = vector_model
        return True
    return False

def get_existing_model(lang: Lang, model: VectorModelType,  corpus: str, dim: int, model_abbr: str) -> VectorModel:
    global EXISTING_MODELS

    if (lang in EXISTING_MODELS and model in EXISTING_MODELS[lang] and corpus in EXISTING_MODELS[lang][model]
        and dim in EXISTING_MODELS[lang][model][corpus]):
        logger.info(f'Model {model_abbr} already exists in memory')
        return EXISTING_MODELS[lang][model][corpus][dim]
    return None

def get_default_model(lang: Lang, check_updates=True) -> VectorModel:
    if lang not in DEFAULT_MODELS:
        return None
    model, corpus = DEFAULT_MODELS[lang]
    return create_vector_model(lang, model, corpus=corpus.value, check_updates=check_updates)

    
def create_vector_model(lang: Lang, model: VectorModelType, corpus: str, dim: int = 300, check_updates=True) -> VectorModel:
    global logger
    model_abbr = f'{lang.value}-{model.name}-{corpus}-{dim}'
    logger.info(f'Getting model {model_abbr}...')
    vector_model = get_existing_model(lang=lang, model=model, corpus=corpus, dim=dim, model_abbr=model_abbr)
    if vector_model is not None:
        return vector_model
    else:
        try:
            if model == VectorModelType.LSA:
                vector_model = LSA(corpus, lang, dim, check_updates=check_updates)
            elif model == VectorModelType.LDA:
                vector_model = LDA(corpus, lang, dim, check_updates=check_updates)
            elif model == VectorModelType.WORD2VEC:
                vector_model = Word2Vec(corpus, lang, dim, check_updates=check_updates)
            elif model == VectorModelType.TRANSFORMER:
                vector_model = TransformersEncoder(lang)
            save_model(vector_model=vector_model, lang=lang, 
                model_type=model, corpus=corpus, dim=dim, model_abbr=model_abbr)
            return vector_model
        except:
            logger.error(f'Model {model_abbr} could not be instantiate.')
        return None



from enum import Enum, unique
 
@unique
class CsclIndicesDescriptions(Enum):
    INTER_ANIMATION_DEGREE = 'Degree of inter-animation'
    SOCIAL_KB = 'Cumulated social KB scores'
    INDEGREE = 'In-degree centrality'
    NO_CONTRIBUTION = 'Contributions'
    CONTRIBUTIONS_SCORE = 'Cumulated contribution scores'
    RHYTHMIC_COEFFICIENT = 'Personal rhythmic coefficient'
    NO_VERBS = 'Verbs'
    FREQ_MAX_RHYTMIC_INDEX = 'Frequency of maximal rhythmic index'
    RHYTHMIC_INDEX = 'Personal rhythmic index'
    PERSONAL_REGULARITY_ENTROPY = 'Personal chat entropy for regularity measure'
    OUTDEGREE = 'Out-degree centrality'
    NO_NOUNS = 'Nouns'




import copy
import csv
import pickle
import random
from enum import Enum, auto
from typing import Dict, List, Tuple

import numpy as np
from rb.complexity.complexity_index import ComplexityIndex
from rb.core.document import Document
from rb.core.lang import Lang


class TargetType(Enum):
    FLOAT = auto()
    INT = auto()
    STR = auto()

class Task:
    def __init__(self, type: TargetType, values: List[str]):
        if type is TargetType.INT:
            self.values = [int(val) for val in values]
            unique = len(set(self.values))
            if max(self.values) - min(self.values) + 1 != unique:
                type = TargetType.STR
        elif type is TargetType.FLOAT:
            self.values = [float(val) for val in values]
            self.min = min(self.values)
            self.max = max(self.values)
        if type is TargetType.STR:
            self.values = values
            self.classes = list({val for val in self.values})
            self.index = {c: i for i, c in enumerate(self.classes)}
        self.type = type
        self.mask: List[bool] = []
        self.train_values = []
        self.dev_values = []

    def _get_targets(self, values) -> List:
        if self.type is TargetType.STR:
            return [self.index[val] for val in values]
        else:
            return values
    
    def get_train_targets(self) -> List:
        return self._get_targets(self.train_values)

    def get_dev_targets(self) -> List:
        return self._get_targets(self.dev_values)

class Dataset:

    def __init__(self, lang: Lang, names: List[str], texts: List[str], labels: List[List[str]]):
        self.lang = lang
        self.names = names
        self.texts = texts
        self.tasks = self.convert_labels(labels)
        self.train_texts: List[str] = []
        self.dev_texts: List[str] = []
        self.features: List[ComplexityIndex] = []
        self.train_features: List[Dict[ComplexityIndex, float]] = []
        self.dev_features: List[Dict[ComplexityIndex, float]] = []
        self.all_features: List[Dict[ComplexityIndex, float]] = []
        self.train_indices: List[int] = []
        self.dev_indices: List[int] = []
    
    def split(self, dev_ratio: float):
        indices = list(range(len(self.texts)))
        random.shuffle(indices)
        n_dev = int(dev_ratio * len(self.texts))
        self.dev_indices = indices[:n_dev]
        self.train_indices = indices[n_dev:]
        self.expand()
    
    def expand(self):
        self.train_texts = [self.texts[index] for index in self.train_indices]
        self.dev_texts = [self.texts[index] for index in self.dev_indices]
        self.train_features = [self.all_features[index] for index in self.train_indices]
        self.dev_features = [self.all_features[index] for index in self.dev_indices]
        for task in self.tasks:
            task.train_values = [task.values[index] for index in self.train_indices]
            task.dev_values = [task.values[index] for index in self.dev_indices]
            task.values = None
    
    def reduce_size(self):
        self.all_features = [
            {
                feature: indices[feature]
                for feature in self.features
            }
            for indices in self.all_features
        ]
        self.train_features = [
            {
                feature: indices[feature]
                for feature in self.features
            }
            for indices in self.train_features
        ]
        self.dev_features = [
            {
                feature: indices[feature]
                for feature in self.features
            }
            for indices in self.dev_features
        ]
    
    def normalize_features(self):
        self.normalized_train_features = [{} for _ in range(len(self.train_features))]
        self.normalized_dev_features = [{} for _ in range(len(self.dev_features))]
        for feature in self.features:
            values = [indices[feature] for indices in self.train_features]
            min_value = min(values)
            max_value = max(values)
            for indices, val in zip(self.normalized_train_features, values):
                indices[feature] = (val - min_value) / (max_value - min_value)
            for indices, val in zip(self.normalized_dev_features, [indices[feature] for indices in self.dev_features]):
                indices[feature] = (val - min_value) / (max_value - min_value)
            
    def convert_labels(self, labels: List[List[str]]) -> List[Task]:
        values = zip(*labels)
        tasks = []
        for targets in values:
            if all(is_double(target) for target in targets):
                tasks.append(Task(TargetType.FLOAT, targets))
            elif all(is_int(target) for target in targets):
                tasks.append(Task(TargetType.INT, targets))
            else:
                tasks.append(Task(TargetType.STR, targets))
        return tasks

    def save_features(self, filename: str):
        with open(filename, "wt", encoding="utf-8") as f:
            writer = csv.writer(f, delimiter=",")
            writer.writerow(["name"] + [repr(index) for index in self.features])
            for name, features in zip(self.names, self.all_features):
                writer.writerow([name] + [features[index] for index in self.features])

    def save(self, filename: str):
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load_features(filename: str) -> List[List[float]]:
        with open(filename, "rt", encoding="utf-8") as f:
            reader = csv.reader(f, delimiter=",")
            header = next(reader)
            return [[float(x) for x in row[1:]] for row in reader]

    @staticmethod
    def load(filename: str) -> "Dataset":
        with open(filename, "rb") as f:
            return pickle.load(f)

def is_double(value: str) -> bool:
    try:
        float(value)
    except ValueError:
        return False
    return True

def is_int(value: str) -> bool:
    try:
        int(value)
    except ValueError:
        return False
    return True



from typing import Dict, List

import numpy as np
import tensorflow as tf
from rb.core.lang import Lang
from rb.processings.encoders.bert import BertWrapper
from rb.processings.pipeline.dataset import Dataset, TargetType, Task
from rb.processings.pipeline.estimator import Classifier, Regressor
from tensorflow import keras
from sklearn.model_selection import KFold, StratifiedKFold
from rb.utils.utils import HiddenPrints

class BertClassifier(Classifier, Regressor):

    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, object], use_indices=True, use_mask=True, shared=True):
        super(Classifier, self).__init__(dataset, tasks, params)
        self.max_seq_length = 256
        self.bert = BertWrapper(dataset.lang, max_seq_len=self.max_seq_length)
        self.tasks = tasks
        self.output = params["output"]
        self.hidden = params["hidden"]
        self.use_indices = use_indices
        self.use_mask = use_mask
        self.shared = shared
        self.initialize()

    def initialize(self):
        tf.keras.backend.clear_session()
        self.model = self.create_model()
        with HiddenPrints():
            self.bert.load_weights()
        self.initial_weights = self.model.get_weights()

    def create_model(self) -> keras.Model:
        inputs, bert_output = self.bert.create_inputs_and_model()
        cls_output = self.bert.get_output(bert_output, self.output)
        features = tf.keras.layers.Input(shape=(len(self.dataset.features),), dtype=tf.float32, name="features")
        outputs = []
        if self.shared:
            global_hidden = tf.keras.layers.Dense(self.hidden, activation="tanh")
        for i, task in enumerate(self.tasks):
            if self.use_indices:
                if self.use_mask:
                    masked_features = keras.layers.Lambda(lambda x: x * task.mask)(features)
                else:
                    masked_features = features
                concat = keras.layers.concatenate([cls_output, masked_features])
            else:
                concat = cls_output
            if self.shared:
                hidden = global_hidden(concat)
            else:
                hidden = tf.keras.layers.Dense(self.hidden, activation="tanh")(concat)
            if task.type is TargetType.FLOAT:
                output = keras.layers.Dense(1, name=f"output{i}")(hidden)
            else:
                output = keras.layers.Dense(len(task.classes), activation='softmax', name=f"output{i}")(hidden)
            outputs.append(output)
        model = keras.Model(inputs=inputs + [features], outputs=outputs)
        optimizer = keras.optimizers.Adam(learning_rate=1e-5)
        losses = {
            f"output{i}": tf.keras.losses.MeanSquaredError() if task.type is TargetType.FLOAT else tf.keras.losses.SparseCategoricalCrossentropy()
            for i, task in enumerate(self.tasks)
        }
        metrics = {
            f"output{i}": tf.keras.metrics.RootMeanSquaredError() if task.type is TargetType.FLOAT else "accuracy"
            for i, task in enumerate(self.tasks)
        }
        model.compile(optimizer=optimizer, loss=losses, metrics=metrics)
        return model
    
    def metric(self, task: Task) -> str:
        return "root_mean_squared_error" if task.type is TargetType.FLOAT else "accuracy"
    
    def cross_validation(self, n=5):
        kf = KFold(n, shuffle=True)
        features = [[indices[feature] for feature in self.dataset.features]
                    for indices in self.dataset.normalized_train_features[:1000]]
        inputs = self.bert.process_input(self.dataset.train_texts[:1000])
        inputs.append(np.array(features))
        outputs = [np.array(task.get_train_targets()[:1000]) for task in self.tasks]
        losses = []
        epochs = []
        for train_index, dev_index in kf.split(inputs[0]):
            train_inputs = [input[train_index] for input in inputs]
            dev_inputs = [input[dev_index] for input in inputs]
            train_outputs = [output[train_index] for output in outputs]
            dev_outputs = [output[dev_index] for output in outputs]
            self.model.set_weights(self.initial_weights)
            history = self.model.fit(train_inputs, train_outputs, batch_size=16, epochs=5, validation_data=[dev_inputs, dev_outputs])
            epoch, loss = min(enumerate(history.history["val_loss"]), key=lambda x: x[1])
            losses.append(loss)
            epochs.append(epoch + 1)
            break
        return int(np.mean(epochs)), np.mean(losses)
            
    def train(self, epochs: int) -> float:
        self.model.set_weights(self.initial_weights)
        train_features = [
            [indices[feature] for feature in self.dataset.features]
            for indices in self.dataset.normalized_train_features[:1000]
        ]
        train_inputs = self.bert.process_input(self.dataset.train_texts[:1000])
        train_inputs.append(np.array(train_features))
        train_outputs = [np.array(task.get_train_targets()[:1000]) for task in self.tasks]
        dev_features = [
            [indices[feature] for feature in self.dataset.features]
            for indices in self.dataset.normalized_dev_features
        ]
        dev_inputs = self.bert.process_input(self.dataset.dev_texts)
        dev_inputs.append(np.array(dev_features))
        dev_outputs = [np.array(task.get_dev_targets()) for task in self.tasks]
        
        history = self.model.fit(train_inputs, train_outputs, batch_size=16, epochs=epochs, validation_data=[dev_inputs, dev_outputs])
        return [history.history[f"val_output{i}_{self.metric(task)}"][-1] for i, task in enumerate(self.tasks)]

    @classmethod
    def parameters(cls):
        return {
            "output": ["cls", "pool"],
            "hidden": [64, 128, 256],
        }
    


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoVerbFormEnum(Enum):
    FIN = auto()
    GER = auto()
    INF = auto()
    PART = auto()


class RoPOSFeatureVerbForm(POSFeature):


    _INSTANCE = None

    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.VBF,
                            values=None, pos_supported=None)
        self.values = set([ro_verb_form for _, ro_verb_form in RoVerbFormEnum.__members__.items()])
        self.pos_supported = set([POS.AUX, POS.VERB, POS.X])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureVerbForm":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureVerbForm()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoVerbFormEnum]:
        pattern = '(VerbForm=[a-zA-Z0-9,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        nrs = matches[0].split('=')[1].split(',')
        values = []
        for nr in nrs:
            if nr == 'Fin':
                values.append(RoVerbFormEnum.FIN)
            elif nr == 'Ger':
                values.append(RoVerbFormEnum.GER)
            elif nr == 'Inf':
                values.append(RoVerbFormEnum.INF)
            elif nr == 'Part':
                values.append(RoVerbFormEnum.PART)
        return values


import tensorflow as tf
import numpy as np

def get_angles(pos, i, d_model):
    angle_rates = 1 / np.power(10000, (2 * (i//2)) / np.float32(d_model))
    return pos * angle_rates

def positional_encoding(position, d_model):
    angle_rads = get_angles(np.arange(position)[:, np.newaxis],
                            np.arange(d_model)[np.newaxis, :],
                            d_model)
    # apply sin to even indices in the array; 2i
    angle_rads[:, 0::2] = np.sin(angle_rads[:, 0::2])
    # apply cos to odd indices in the array; 2i+1
    angle_rads[:, 1::2] = np.cos(angle_rads[:, 1::2])
    pos_encoding = angle_rads[np.newaxis, ...]
        
    return tf.cast(pos_encoding, dtype=tf.float32)

def create_padding_mask(seq):
    seq = tf.cast(tf.math.equal(seq, 0), tf.float32) # returns 0 and 1 float
    # add extra dimensions to add the padding
    # to the attention logits.
    return seq[:, tf.newaxis, tf.newaxis, :]  # (batch_size, 1, 1, seq_len)

# def create_padding_mask_np(seq):
#     seq = np.equal(seq, 0).astype(dtype=np.float32)
#     return seq[:, np.newaixs, np.newaxis, :]

def create_look_ahead_mask(size):
    mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0) # select band diagonal -> lower this case
    return mask  # (seq_len, seq_len)

# def create_look_ahead_mask_np(size):
#     mask = 1 - np.tril(np.ones(size, size), 0)
#     return mask  # (seq_len, seq_len)

def point_wise_feed_forward_network(d_model, dff):
    """point wise nn because if u have (seq_length, hidden) at each seq_length you multiple 
        by a the same dense layer. this is the default behaviour of a dense layer in keras anyways """
    return tf.keras.Sequential([
        tf.keras.layers.Dense(dff, activation='relu'),  # (batch_size, seq_len, dff)
        tf.keras.layers.Dense(d_model)  # (batch_size, seq_len, d_model)
    ])

def create_masks(inp, tar):
    # Encoder padding mask
    enc_padding_mask = create_padding_mask(inp)
    
    # Used in the 2nd attention block in the decoder.
    # This padding mask is used to mask the encoder outputs.
    dec_padding_mask = create_padding_mask(inp)
    
    # Used in the 1st attention block in the decoder.
    # It is used to pad and mask future tokens in the input received by 
    # the decoder.
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)
    
    return enc_padding_mask, combined_mask, dec_padding_mask






from rb.core.word import Word
from rb.comprehension.utils.graph.cm_node_type import CmNodeType


class CmNodeDO:

    def __init__(self, word: Word, node_type: CmNodeType) -> None:
        self.word = word
        self.node_type = node_type
        self.active = False
        self.activation_score = 0.0

    def get_word(self) -> Word:
        return self.word
    

    def get_node_type(self) -> CmNodeType:
        return self.node_type


    def set_node_type(self, node_type: CmNodeType) -> None:
        self.node_type = node_type

    
    def is_active(self) -> bool:
        return self.active


    def activate(self) -> None:
        self.active = True


    def deactivate(self) -> None:
        self.active = False


    def get_activation_score(self) -> float:
        return self.activation_score 


    def set_activation_score(self, activation_score: float) -> None:
        self.activation_score = activation_score

    
    def increment_activation_score(self) -> None:
        self.activation_score += 1


    def __repr__(self):
        return "CmNodeDO(%r, %r, %r, %r)" % (self.word.lemma, self.node_type, self.activation_score, self.active)


    def __str__(self):
        return "CmNodeDO(%r, %r, %r, %r)" % (self.word.lemma, self.node_type, self.activation_score, self.active)

    
    def __eq__(self, other):
        if isinstance(other, CmNodeDO):
            return self.word.lemma == other.word.lemma
        return NotImplemented


    def __hash__(self):
        return hash(tuple(self.word.lemma))




from typing import Dict, List, Tuple, Union

import numpy as np
import tensorflow as tf
from networkx import Graph
from networkx.algorithms.components.connected import connected_components
from rb import POS, Document, Lang, Word
from rb.utils.downloader import check_version, download_model
from transformers import AutoConfig, AutoTokenizer, TFAutoModelForMaskedLM


class ChainsModel():

    def __init__(self, lang = Lang.EN, linear=False, check_updates=True):
        if lang is not Lang.EN:
            raise NotImplementedError(f"Lang {lang.value} not available")
        self.lang = lang
        self.linear = linear
        self.model_type = "linear" if linear else "mlp"
        if check_updates and check_version(self.lang, ["models", "chains", self.model_type]):
            download_model(lang, ["models", "chains", self.model_type])
        self.build_model()
        model_name = "bert-base-cased"
        self.layers = 12
        self.heads = 12
        conf = AutoConfig.from_pretrained(model_name, output_attentions=True, output_past=False, output_additional_info=False, output_hidden_states=False)
        self.tokenizer = AutoTokenizer.from_pretrained(model_name)
        self.bert_model = TFAutoModelForMaskedLM.from_pretrained(model_name, config=conf)
        self.max_seq_len = 256
    
    def build_model(self):
        input = tf.keras.layers.Input(288, dtype=tf.float32)
        if self.linear:
            output = tf.keras.layers.Dense(1, activation="sigmoid")(input)
        else:
            hidden = tf.keras.layers.Dense(128, activation="relu")(input)
            hidden = tf.keras.layers.Dense(64, activation="relu")(hidden)
            output = tf.keras.layers.Dense(1, activation="sigmoid")(hidden)
        self.model = tf.keras.Model(inputs=[input], outputs=[output])
        self.model.load_weights(f"resources/{self.lang.value}/models/chains/{self.model_type}/{self.model_type}")

    @staticmethod
    def common_prefix(word1: str, word2: str) -> str:
        i = 0
        while i < min(len(word1), len(word2)):
            if word1[i] != word2[i]:
                return word1[:i]
            i += 1
        return word1[:i]

    def create_word_token_dict(self, doc: Document, tokens: List[int]) -> Dict[Word, List[int]]:
        i = 1
        result = {}
        tokens = [self.tokenizer._convert_id_to_token(token).replace("##", "") for token in tokens]
        current = ""
        for word in doc.get_words():
            if len(word.text) == 0:
                continue
            while i < len(tokens) - 1:
                current = self.common_prefix(word.text, tokens[i])
                if len(current) > 0:
                    break
                i += 1
            ids = [i]
            if word.text == current:
                if len(word.text) < len(tokens[i]):
                    tokens[i] = tokens[i][len(current):]
                else:
                    i += 1
            else: 
                i += 1
                while i < len(tokens) and word.text.startswith(current + tokens[i]):
                    ids.append(i)
                    current += tokens[i]
                    i += 1
                current = word.text[len(current):]
                if len(current) > 0:
                    tokens[i] = tokens[i][len(current):]
            if word.pos in {POS.PRON, POS.NOUN, POS.VERB, POS.PROPN} and not word.is_modal_verb():
                result[word] = ids
        return result
    
    def build_similarity_matrix(self, doc: Union[str, Document]) -> Tuple[List[Word], np.ndarray]:
        if isinstance(doc, str):
            doc = Document(self.lang, doc)
        inputs = self.tokenizer(doc.text)
        start = 1
        n = len(inputs["input_ids"]) - 1
        word_dict = self.create_word_token_dict(doc, inputs["input_ids"])
        words = list(word_dict.keys())
        words_index = {w: i for i, w in enumerate(words)}
        word_scores = np.zeros([len(words), len(words), self.layers, self.heads], np.float32)
        while start < n:
            end = min(n, start + self.max_seq_len - 2)
            input_ids = tf.convert_to_tensor(np.array([[101] + inputs["input_ids"][start:end] + [102]]))
            attention_mask = tf.convert_to_tensor(np.ones(input_ids.shape, dtype=np.int32))
            outputs = self.bert_model(input_ids, attention_mask=attention_mask)
            words_in_batch = [
                (word, [id - start for id in ids if start <= id < end]) 
                for word, ids in word_dict.items() 
                if any(id for id in ids if start <= id < end)
            ]
            scores = np.concatenate(outputs[1], axis=0)[:, :, 1:-1, 1:-1]
            for wb1 in words_in_batch:
                w1 = words_index[wb1[0]]
                for wb2 in words_in_batch:
                    score = sum(scores[:, :, id2, id1] for id1 in wb1[1] for id2 in wb2[1]) / len(wb2[1])
                    w2 = words_index[wb2[0]]
                    for l in range(self.layers):
                        for h in range(self.heads):
                            word_scores[w1, w2, l, h] = max(word_scores[w1, w2, l, h], score[l, h])
            start += self.max_seq_len // 2
        all_features = []
        all_pairs = []
        for i in range(len(words) - 1):
            for j in range(i+1, min(i + self.max_seq_len // 2, len(words))):
                features = [word_scores[i, j, l, h] for l in range(self.layers) for h in range(self.heads)]
                features += [word_scores[j, i, l, h] for l in range(self.layers) for h in range(self.heads)]
                if sum(features) != 0:
                    all_features.append(np.array(features))
                    all_pairs.append((i, j))
        pred_scores = self.model.predict(np.array(all_features))     
        scores = np.zeros([len(words), len(words)], dtype=np.float32)
        for pair, score in zip(all_pairs, pred_scores):
            i, j = pair
            scores[i, j] = score[0]
        return words, scores
    
    def build_semantic_chains(self, doc: Union[str, Document], threshold=0.9) -> List[List[Word]]:
        words, similarity = self.build_similarity_matrix(doc)
        graph = Graph()
        for i, word1 in enumerate(words):
            for j in range(i+1, len(words)):
                if similarity[i, j] > threshold:
                    graph.add_edge(word1, words[j])
        
        return [
            list(sorted(component, key=lambda w: w.index_in_doc)) 
            for component in connected_components(graph)
            ]


        
    
        
		


from rb.core.lang import Lang
from rb.similarity.vector_model import VectorModel, VectorModelType
from rb.similarity.vector_model_factory import create_vector_model
from rb.core.word import Word
import rb.similarity.wordnet as wordnet
from typing import List
from typing import Dict

def create_semantic_models(lang: Lang) -> List[VectorModel]:
    if lang is Lang.RO:
        return [create_vector_model(Lang.RO, VectorModelType.from_str('word2vec'), "readme")]
    if lang is Lang.EN:
        semantic_models = []
        semantic_models.append(create_vector_model(Lang.EN, VectorModelType.from_str("word2vec"), "coca"))
        # semantic_models.append(create_vector_model(Lang.EN, VectorModelType.from_str("lsa"), "coca"))
        # semantic_models.append(create_vector_model(Lang.EN, VectorModelType.from_str("lda"), "coca"))
        return semantic_models
    else:
        return None


def avg_similarity_using_models(word1: Word, word2: Word, semantic_models: List[VectorModel]) -> float:
        avg_models_similarity = 0.0
        for sm in semantic_models:
            avg_models_similarity += sm.similarity(word1, word2)
        if semantic_models:
            avg_models_similarity /= len(semantic_models)
        return avg_models_similarity


def dictify_similar_concepts(word: Word, lang: Lang, similar_concepts: List[str], semantic_models: List[VectorModel]) -> Dict:
    data = {}
    concepts = {}
    for concept in similar_concepts:
        concept_word = Word.from_str(lang, concept)
        concepts["{}({}, null, null)".format(concept, concept_word.lemma)] = avg_similarity_using_models(word, concept_word, semantic_models)
    
    data["concepts"] = concepts

    return {"data": data}


def get_similar_concepts(raw_word: str, lang: Lang) -> List[str]:
    semantic_models = create_semantic_models(lang)
    if not semantic_models:
        return []
    
    word = Word.from_str(lang, raw_word)
    synonyms = wordnet.get_synonyms(word)
    hypernyms = wordnet.get_hypernyms(word)
    similar_concepts = []
    similar_concepts.extend(synonyms)
    similar_concepts.extend(hypernyms)

    for vect_model in semantic_models:
        closest_semantic_words = vect_model.most_similar(word, topN=5, threshold=0.5)
    similar_concepts.extend([x[0] for x in closest_semantic_words])
    similar_concepts = list(set(similar_concepts))
    # remove the word if that is the case
    similar_concepts = [x for x in similar_concepts if x != word]

    response = dictify_similar_concepts(word, lang, similar_concepts, semantic_models)
    response["success"] = True
    response["errorMsg"] = ""

    return response